#!/usr/bin/env luajit
-- aomodel.lua
-- (C) 2013 Lesley De Cruz
-- See LICENSE.txt for license information.

------------------------------------------------------------------------
-- Lua implementation of the 24-variable low-order coupled
-- ocean-atmosphere model OA-QG-WS v2.
--
-- Optimised for LuaJIT (http://luajit.org).
--
-- See @{readme.md} for more details.
------------------------------------------------------------------------

local n = 24 + 1 -- 24 variables + constant term
local params = require("params")

-- Utilities
local array = require("array")(n) -- n-array constructor
local Heun = require("rk2")(n) -- integrator
local st = require("stat")(n) -- statistics object

-- IO functions
local format = string.format
local function fprintf(f,...) return f:write(format(...)) end
local function write(t,y,f)
  f = f or io.stdout
  f:write(t)
  for i=1,n-1 do fprintf(f, "\t%.12f",y[i]) end
  f:write("\n")
end
local function close(f) if f and f~=io.stdout then f:close() end end

-- Tensor with the coefficients of the nonlinear (polynomial) system of
-- diffeqs. N.B.: the model tensor is fixed, so changes to model parameters
-- will only be effective up to this point!
local tensor = require("tensor")
local aotensor = tensor.coo_to_fficoo(tensor.simplify_coo(require("aotensor")))
local sparse_mul3 = tensor.sparse_mul3

------------------------------------------------------------------------
-- Integrate coupled ocean-atmosphere model.
------------------------------------------------------------------------

--- Function that calculates the time derivative of the 24 variables.
-- The function reduces to a sparse tensor contraction due to the bilinear
-- nature of the equations.
-- @param t time
-- @param y array with variables at time t
-- @param buf n-arrayi (buffer) to store derivatives.
local function ao(t,y,buf)
  return sparse_mul3(aotensor,y,y,buf)
end


-- Get initial conditions (for this value of thetastar)
local y
do
  local IC = require("IC")
  local initialconditions = IC[params.m.thetastar] or IC[0.10]
  assert(#initialconditions==n-1)
  initialconditions[0] = 1
  y = array({data=initialconditions})
end

-- Perform integration (and write out).
local dt = params.i.dt
for i=0,params.i.ntrans-1 do
  Heun(y,ao,i*dt,dt,y)
end

for i=1,n do
  assert(y[i]==y[i], "NaN encountered; the system is unstable. Please modify your initial conditions, heating or coupling parameter.")
end

local logf = params.i.writeout and io.open(params.i.getoutfn(),"w") or io.stdout
for i=0,params.i.niter-1 do
  if i%(10/dt)==0 then
    if params.i.writeout then write(i*dt,y,logf) end
    st.acc(y)
  end
  Heun(y,ao,i*dt,dt,y)
end
if params.i.writeout then write(params.i.niter*dt,y,logf) end
close(logf)

-- Write out statistics
logf = io.open(params.i.getmeanfn(),"w") or io.stdout
write("mean", st.mean(),logf)
write("var", st.var(),logf)
fprintf(logf,"samples\t %d\n", st.iter())
close(logf)
