-- aotensor.lua
-- (C) 2013 Lesley De Cruz
-- See LICENSE.txt for license information.

------------------------------------------------------------------------
-- The equation tensor for the coupled ocean-atmosphere model.
------------------------------------------------------------------------

-- Model parameters and the inner products between the truncated fields.
local m = require("params").m
local inprod = require("inprod")

-- Translate field names into effective coordinates
local function psi(i) return i end
local function theta(i) return i+10 end
local function A(i) return i+20 end

--- Kronecker delta
local function kdelta(i,j) return i==j and 1 or 0 end

--- Model tensor. The tensor (in coordinate list form) that holds the
-- coefficients for the bilinear system of differential equations.
-- Constructed as an upper diagonal tensor (in the last two dimensions).
local coo_list = {}

--- Coefficient insertion function.
-- Creates an upper diagonal tensor (in the last two dimensions)
-- i.e. {i,k,j,v} will be stored as {i,j,k,v} if j<=k.
local function coeff(i,j,k,v)
  if v and v~=0 then
    coo_list[#coo_list+1]= j<=k and {i,j,k,v} or {i,k,j,v}
  end
end

-- Atmosphere equations
------------------------------------------------------------------------

local atmos = inprod.atmos
local a, b, c, d, g = atmos.a, atmos.b, atmos.c, atmos.d, atmos.g
local kd, betp = m.kd, m.betp
local hpp, thetastar, kdp, sig0 = m.hpp, m.thetastar, m.kdp, m.sig0

coeff(theta(1),0,0, (hpp*thetastar)/(1 - a[1][1]*sig0)) -- constant forcing
-- Contributions proportional to...
for i=1,10 do
  for j=1,10 do
    -- psi_j
    coeff(psi(i),psi(j),0,-((c[i][j]*betp)/a[i][i]) - (kd*kdelta(i, j))/2)
    coeff(theta(i),psi(j),0, (a[i][j]*kd*sig0)/(-2 + 2*a[i][i]*sig0))
    -- theta_j
    coeff(psi(i),theta(j),0, (kd*kdelta(i, j))/2)
    coeff(theta(i),theta(j),0, (-(sig0*(2*c[i][j]*betp + a[i][j]*(kd + 4*kdp))) + 2*hpp*kdelta(i, j))/(-2 + 2*a[i][i]*sig0))
    for k=1,10 do
      -- psi_j x psi_k
      coeff(psi(i),psi(j),psi(k), -(b[i][j][k]/a[i][i]))
      -- theta_j x theta_k
      coeff(psi(i),theta(j),theta(k), -(b[i][j][k]/a[i][i]))
      -- psi_j x theta_k
      coeff(theta(i),psi(j),theta(k), (g[i][j][k] - b[i][j][k]*sig0)/(-1 + a[i][i]*sig0))
      -- theta_j x psi_k
      coeff(theta(i),theta(j),psi(k), (b[i][j][k]*sig0)/(1 - a[i][i]*sig0))
    end
  end
  for j=1,4 do -- A_j
    coeff(psi(i),A(j),0,  kd*d[i][j]/(2*a[i][i]))
    coeff(theta(i),A(j),0, kd*(d[i][j]*sig0)/(2 - 2*a[i][i]*sig0))
  end
end

-- Ocean equations
------------------------------------------------------------------------

local ocean = inprod.ocean
local C, K, M, N = ocean.C, ocean.K, ocean.M, ocean.N
local dp, rp, G = m.dp, m.rp, m.G

-- Contributions propoportional to...
for j=1,10 do  -- psi_j - theta_j == psi3_j
  coeff(A(1),psi(j),0,-((dp*(-(K[3][j]*M[1][3]) + K[1][j]*(G + M[3][3])))/
    (M[1][3]*M[3][1] - (G + M[1][1])*(G + M[3][3]))))
  coeff(A(2),psi(j),0,-((dp*(-(K[4][j]*M[2][4]) + K[2][j]*(G + M[4][4])))/
    (M[2][4]*M[4][2] - (G + M[2][2])*(G + M[4][4]))))
  coeff(A(3),psi(j),0,(dp*K[3][j]*(G + M[1][1]) - dp*K[1][j]*M[3][1])/
    (-(M[1][3]*M[3][1]) + (G + M[1][1])*(G + M[3][3])))
  coeff(A(4),psi(j),0,(dp*K[4][j]*(G + M[2][2]) - dp*K[2][j]*M[4][2])/
    (-(M[2][4]*M[4][2]) + (G + M[2][2])*(G + M[4][4])))
  coeff(A(1),theta(j),0,((dp*(-(K[3][j]*M[1][3]) + K[1][j]*(G + M[3][3])))/
    (M[1][3]*M[3][1] - (G + M[1][1])*(G + M[3][3]))))
  coeff(A(2),theta(j),0,((dp*(-(K[4][j]*M[2][4]) + K[2][j]*(G + M[4][4])))/
    (M[2][4]*M[4][2] - (G + M[2][2])*(G + M[4][4]))))
  coeff(A(3),theta(j),0,-(dp*K[3][j]*(G + M[1][1]) - dp*K[1][j]*M[3][1])/
    (-(M[1][3]*M[3][1]) + (G + M[1][1])*(G + M[3][3])))
  coeff(A(4),theta(j),0,-(dp*K[4][j]*(G + M[2][2]) - dp*K[2][j]*M[4][2])/
    (-(M[2][4]*M[4][2]) + (G + M[2][2])*(G + M[4][4])))
end
for j=1,4 do   -- A_j
  coeff(A(1),A(j),0,
    (-(((rp + dp)*M[3][j] + betp*N[3][j])*M[1][3]) + (rp + dp)*M[1][j]*(G + M[3][3]) + betp*N[1][j]*(G + M[3][3]))/
    (M[1][3]*M[3][1] - (G + M[1][1])*(G + M[3][3])))
  coeff(A(2),A(j),0,
    (-(((rp + dp)*M[4][j] + betp*N[4][j])*M[2][4]) + (rp + dp)*M[2][j]*(G + M[4][4]) + betp*N[2][j]*(G + M[4][4]))/
    (M[2][4]*M[4][2] - (G + M[2][2])*(G + M[4][4])))
  coeff(A(3),A(j),0,
    (-(((rp + dp)*M[3][j] + betp*N[3][j])*(G + M[1][1])) + ((rp + dp)*M[1][j] + betp*N[1][j])*M[3][1])/
    (-(M[1][3]*M[3][1]) + (G + M[1][1])*(G + M[3][3])))
  coeff(A(4),A(j),0,
    (-(((rp + dp)*M[4][j] + betp*N[4][j])*(G + M[2][2])) + ((rp + dp)*M[2][j] + betp*N[2][j])*M[4][2])/
    (-(M[2][4]*M[4][2]) + (G + M[2][2])*(G + M[4][4])))
  for k=1,4 do -- A_j x A_k
    coeff(A(1),A(j),A(k),(-(C[3][j][k]*M[1][3]) + C[1][j][k]*(G + M[3][3]))/
      (M[1][3]*M[3][1] - (G + M[1][1])*(G + M[3][3])))
    coeff(A(2),A(j),A(k),(-(C[4][j][k]*M[2][4]) + C[2][j][k]*(G + M[4][4]))/
      (M[2][4]*M[4][2] - (G + M[2][2])*(G + M[4][4])))
    coeff(A(3),A(j),A(k),(-(C[3][j][k]*(G + M[1][1])) + C[1][j][k]*M[3][1])/
      (-(M[1][3]*M[3][1]) + (G + M[1][1])*(G + M[3][3])))
    coeff(A(4),A(j),A(k),(-(C[4][j][k]*(G + M[2][2])) + C[2][j][k]*M[4][2])/
      (-(M[2][4]*M[4][2]) + (G + M[2][2])*(G + M[4][4])))
  end
end

-- @export
return coo_list
