-- params.lua
-- (C) 2013 Lesley De Cruz
-- See LICENSE.txt for license information.

------------------------------------------------------------------------
-- Integration parameters and physical parameters for the coupled
-- ocean-atmosphere model.
------------------------------------------------------------------------

local sqrt, cos, sin, pi = math.sqrt, math.cos, math.sin, math.pi

--- Integration parameters
local i = {
  ntrans  = 1e6,  -- transient period (e.g. 1e7)
  niter   = 1e7,  -- length of trajectory on the attractor (e.g. 5e8)
  dt      = 1e-2, -- the time step
  out_prefix = "aomodel_", -- prefix for the output file name
  writeout = false, -- write out all variables every 10 time units
}

--- Model parameters
local m = {
  --- Scale parameters for the ocean and the atmosphere
  scale = 5e6,     -- the characteristic space scale, L*pi
  f0    = 1.032e-4, -- Coriolis parameter at 45 degrees latitude
  n     = 1.5,     -- aspect ratio (n = 2Ly/Lx ; Lx = 2*pi*L/n; Ly = pi*L)
  rra   = 6370e3,  -- earth radius
  phi0  = pi/4,    -- latitude
  --- Parameters for the ocean
  gp    = 3.1e-2,  -- reduced gravity
  r     = 1e-8,    -- frictional coefficient at the bottom of the ocean
  H     = 5e2,     -- depth of the water layer
  alpha = 1,       -- coefficient characterizing the intensification of the flow on the western boundaries
  d     = 2e-7,    -- the coupling parameter (should be divided by f0 in order to be adimensional)
  -- Parameters for the atmosphere
  k    = 2e-2, -- bottom friction coefficient
  kp   = 4e-2, -- internal friction coefficient
  hpp  = 4e-2, -- coefficient associated with the thermal relaxation time scale (Newtonian heating)
  sig0 = 1e-1, -- static stability
  thetastar = 0.10 -- Newtonian heating parameter; a value of 0.14 should produce chaos.
}

-- Derived/nondimensionalized quantities
m.L    = m.scale/pi -- characteristic length scale divided by pi
m.LR   = sqrt(m.gp*m.H)/m.f0 -- reduced Rossby deformation radius
m.G    = -m.L^2/m.LR^2    -- \gamma
m.betp = m.L/m.rra*cos(m.phi0)/sin(m.phi0) -- \beta prime
m.rp   = m.r/m.f0         -- r \prime
m.dp   = m.d/m.f0         -- \delta \prime
m.kd   = m.k*2            -- bottom friction coefficient
m.kdp  = m.kp             -- internal friction coefficient
m.timeunit = 1./(m.f0*24.*3600.) -- dimensional time unit

--- output file name for the trajectory
function i.getoutfn()
  return string.format("%sths%.4f_dp%.4e.txt",i.out_prefix,m.thetastar,m.dp)
end
--- output file name for statistics
function i.getmeanfn()
  return string.format("%sths%.4f_dp%.4e_meanfields.txt",i.out_prefix,m.thetastar,m.dp)
end

return {i=i, m=m}

