-- tensor.lua
-- (C) 2013 Lesley De Cruz
-- See LICENSE.txt for license information.

------------------------------------------------------------------------
-- Sparse tensor tools.
--
-- Tensors are entered as coordinate lists:
--    { {x1,y1,z1,val1}, {x2,y2,z2,val2},...}
--
-- This module provides functions to convert a coordinate lists to a nested
-- sparse table or to a (non-growable) fast ffi coordinate list.
--
-- While not the most (space-)efficient sparse tensor storage method, it's
-- simple, generic (for variable dimensions) and fast enough for
-- contractions with 1D arrays.
------------------------------------------------------------------------

local ffi = require("ffi")
local ipairs, tonumber, assert, yield = ipairs, tonumber, assert, coroutine.yield
if not DEBUG then assert = function(v) return v end end

--- Convert a coordinate list of arbitrary dimension to a sparse nested table.
-- Input format: { {x1,y1,z1,val1}, {x2,y2,z2,val2},...}
-- In the returned table, we have
--     t[x1][y1][z1]==val1
-- e.g.
--     mycoolist = {{1, 5, 0.1},{2, 3, 3.14}}
-- yields
--     {[1]={[5]=0.1},[2]={[3]=3.14}}
-- Missing entries are treated as zero in mathematical operations.
-- @function coo_to_sparse_table
-- @param coo_list coordinate list (Lua table)
-- @return sparse table
local coo_to_sparse_table
do
  local sparse_mt
  local function asnumber(x) return tonumber(x) or 0 end
  sparse_mt = {
    -- Automagically generate tables.
    __index = function(t,k,v) if tonumber(k) then return setmetatable({},sparse_mt) end end,
    -- Nonexistent entries behave as zero in mathematical operations.
    __add = function(a,b) return asnumber(a) + asnumber(b) end,
    __sub = function(a,b) return asnumber(a) - asnumber(b) end,
    __unm = function(a) return 0 end,
    __mul = function(a,b) return asnumber(a) * asnumber(b) end,
    __div = function(a,b) return asnumber(a) / asnumber(b) end,
  }
  -- Recursively insert a value dim coordinates deep into a table.
  local function insert_nested(t,level,dim,coo)
    if level==dim then
      -- If more than one entry at a given set of coordinates, take the sum.
      t[coo[level]] = t[coo[level]] + coo[dim+1]
      return
    end
    t[coo[level]] = t[coo[level]] or setmetatable({},sparse_mt)
    insert_nested(t[coo[level]],level+1,dim,coo)
  end
  function coo_to_sparse_table(coo_list)
    local t = setmetatable({},sparse_mt)
    local dim = #coo_list[1]-1
    for _,coo in ipairs(coo_list) do
      insert_nested(t,1,dim,coo)
    end
    return t
  end
end

-- Get the next entry from a nested sparse table in coordinate list form:
-- val1, z1, y1, x1
-- Note that the entries are reversed!
local function next_coo(t,...)
  if type(t)~="table" then yield(t,...)
  else
    for k,t2 in pairs(t) do next_coo(t2,k,...) end
  end
end

-- Reverse the entries of a table.
local function treverse(t)
  for k=1,#t/2 do
    t[k],t[#t+1-k]=t[#t+1-k],t[k]
  end
  return t
end

--- Convert a sparse tensor encoded as a nested table into a coordinate list.
-- @param sparse_table sparse, nested table
-- @return coordinate list (Lua table)
local function sparse_table_to_coo(sparse_table)
  local getnextcoo = coroutine.wrap(function() return next_coo(sparse_table) end)
  local coo_list = {}
  while true do
    local coo = treverse{getnextcoo()}
    if coo[1] then coo_list[#coo_list+1] = coo
    else break end
  end
  return coo_list
end

--- Merge duplicate entries in a coordinate list.
-- This is done by converting it to a sparse table and back.
-- @param coo_list coordinate list (Lua table)
-- @return simplified coordinate list (Lua table)
local function simplify_coo(coo_list)
  return sparse_table_to_coo(coo_to_sparse_table(coo_list))
end

--- Convert a coordinate list of arbitrary dimension to an ffi coordinate list.
-- Input format: { {x1,y1,z1,val1}, {x2,y2,z2,val2},...}
-- e.g.
--     mycoolist = {{1, 5, 0.1}, {2, 3, 3.14}, {3, 2, 10}}
-- yields a struct fficoo which has
--     fficoo.dim == 2
--     fficoo.nelem == 3
--     fficoo.data[0].c[0] == 1
--     fficoo.data[0].c[1] == 5
--     fficoo.data[0].v == 0.1
--     fficoo.data[1].c[0] == 2
--  etc.
-- @param coo_list coordinate list (Lua table)
-- @return ffi coordinate list (ffi struct)
local function coo_to_fficoo(coo_list)
  local dim, nelem = #coo_list[1]-1, #coo_list
  local coo_entry_t = ffi.typeof("struct {double v; int32_t c[$];}",dim)
  local fficoo_t = ffi.typeof("struct{ int32_t dim,nelem; $ data[$];}",coo_entry_t,nelem)
  local inittab = {dim=dim, nelem=nelem, data={}} -- table used to initialize the fficoo struct.
  for i,coo in ipairs(coo_list) do
    local init_coo = {v=coo[dim+1], c={}}
    for j=1,dim do init_coo.c[j-1]=coo[j] end
    inittab.data[i] = init_coo
  end
  return fficoo_t(inittab)
end

--- Sparse multiplication of three tensors, C[i][j][k] a[j] b[k].
-- Note that it is NOT safe to pass `arr_j`/`arr_k` as a result buffer, 
-- as this operation does multiple passes.
-- @param fficoo_ijk an ffi coordinate list (ffi struct) of which index
-- 2 and 3 will be contracted.
-- @param arr_j the array to be contracted with index 2 of ffi_coo_ijk
-- @param arr_k the array to be contracted with index 3 of ffi_coo_ijk
-- @param res array (buffer) to store the result of the contraction
-- @return the result array
local function sparse_mul3(fficoo_ijk, arr_j, arr_k, res)
  res:clear()
  local dim = fficoo_ijk.dim
  assert(dim==3, "sparse_mul3 expects a coordinate list with exactly 3 coordinates")
  assert(fficoo_ijk.data[1].v, "sparse_mul3 expects an ffi coordinate list.")
  for ci = 0,fficoo_ijk.nelem-1 do -- sum over j and k
    local coo = fficoo_ijk.data[ci]
    res.data[coo.c[0]] = res.data[coo.c[0]] + coo.v*arr_j.data[coo.c[1]]*arr_k.data[coo.c[2]]
  end
  return res
end

--- @export
return {
  coo_to_sparse_table = coo_to_sparse_table,
  sparse_table_to_coo = sparse_table_to_coo,
  simplify_coo = simplify_coo,
  coo_to_fficoo = coo_to_fficoo,
  sparse_mul3 = sparse_mul3,
}
